/*
  CodingDrone Rev.0 by DomeKit (BNC)
*/

#include <avr/io.h>

#include "Arduino.h"
#include "BNC_CodingDrone.h"

enum Drone_State{
    STATE_IDLE,
    COMM_READY,
    READY_TO_TAKEOFF,
    IN_FLIGHT,
    FAIL_SAFE,
    LANDING,
    LOWBAT
  };

//// CodingDrone Class Function

CodingDrone::CodingDrone() {
	
}


void CodingDrone::begin()
{
	
	Direction.Up = 'U';
	Direction.Down = 'D';
	Direction.Left = 'L';
	Direction.Right = 'R';
	Direction.Forward = 'F';
	Direction.Backward = 'B';
	Direction.CW_Rotation = 'C';
	Direction.CCW_Rotation = 'W';
	Direction.None = 'N';
	
	
	System.TakeOff = 'T';
	System.Landing = 'N';
	System.Emergency = 'E';
	System.None = 'N';
	
	sensorData = 0;
	Drone_motor_Flag = 0;
	
	sensorCalibrationSwitch = false;
	
	uint8_t i = 0;
	uint8_t j = 0;
	for(i = 0;i < 8; i++){
		motor[i] = 0;
		rcData[i] = 1500;
		cmd[i] = 1500;
	}
	
	for(i = 0;i < 4; i++){
		for(j = 0;j < 4; j++){
			motor_Data[i][j] = 0;
		}
	}	
	
	cmd_sel[0] = Direction.Left;
	cmd_sel[1] = Direction.Right;
	cmd_sel[2] = Direction.Forward;
	cmd_sel[3] = Direction.Backward;
	cmd_sel[4] = Direction.CCW_Rotation;
	cmd_sel[5] = Direction.CW_Rotation;
	cmd_sel[6] = Direction.Up;
	cmd_sel[7] = Direction.Down;
	cmd_sel[8] = System.TakeOff;
	cmd_sel[9] = System.Landing;
	
	Drone_State_Flag = STATE_IDLE;
	read_CMD_Done = 0;
	
	//Serial.begin(115200);
	MODE = 0;
	take_off_height = 100;
	rcSerialCount = 80;
	rcUnCorrectCount = 0;
	
	Pin_Motor_FL = 0; // FL : Front Left 전방-왼쪽
	Pin_Motor_FR = 0; // FR : Front Right 전방-오른쪽
	Pin_Motor_RL = 0; // RL : Rear Left 후방-왼쪽
	Pin_Motor_RR = 0; // RR : Rear Right 후방-오른쪽
	
	
  //블루투스 통신을 초기화 합니다.
  //블루투스 통신속도는 기본적으로 115200입니다.
  Initialize_BT();
  
  //출력핀을 초기화 합니다.
  Initialize_InOutPins();
  
  
  //드론 세팅값을 불러옵니다.
  init_Drone_Setting();
  
  // I2C 톹신 / 가속도 자이로 센서(MPU6050) / 대기압센서 초기화
  // 일반적으로 아두이노에서 사용하는 Wire() 대신 코딩드론 전용 함수를 사용합니다.
  initSensors();
  
  //Drone_Motor_1 = 0;
  //Drone_Motor_2 = 0;
  //Drone_Motor_3 = 0;
  //Drone_Motor_4 = 0;
  
  //Serial.begin(115200);
}

void CodingDrone::MotorSetting(int M_1, int M_2, int M_3, int M_4){
	if(Drone_motor_Flag != 1){
		Pin_Motor_FL = M_1; // FL : Front Left 전방-왼쪽
		Pin_Motor_FR = M_2; // FR : Front Right 전방-오른쪽
		Pin_Motor_RL = M_3; // RL : Rear Left 후방-왼쪽
		Pin_Motor_RR = M_4; // RR : Rear Right 후방-오른쪽
		
		Drone_motor_Flag = 1;
		
		pinMode(Pin_Motor_FL, OUTPUT);
		pinMode(Pin_Motor_FR, OUTPUT);
		pinMode(Pin_Motor_RL, OUTPUT);
		pinMode(Pin_Motor_RR, OUTPUT);
		
		//uint8_t addr[4] = {Motor_select(Pin_Motor_FL),Motor_select(Pin_Motor_FR),Motor_select(Pin_Motor_RL),Motor_select(Pin_Motor_RR)};
		motor_addr[0] = Motor_select(Pin_Motor_FL);
		motor_addr[1] = Motor_select(Pin_Motor_FR);
		motor_addr[2] = Motor_select(Pin_Motor_RL);
		motor_addr[3] = Motor_select(Pin_Motor_RR);
	}
}


// 드론 모터제어 명령처리
void CodingDrone::Motor_LeftRight(int speed_1, int speed_2, int speed_3, int speed_4) {
	motor_Data[0][0] = constrain(speed_1,-50,50);
	motor_Data[0][1] = constrain(speed_2,-50,50);
	motor_Data[0][2] = constrain(speed_3,-50,50);
	motor_Data[0][3] = constrain(speed_4,-50,50);
}

void CodingDrone::Motor_ForwardBackward(int speed_1, int speed_2, int speed_3, int speed_4) {
	motor_Data[1][0] = constrain(speed_1,-50,50);
	motor_Data[1][1] = constrain(speed_2,-50,50);
	motor_Data[1][2] = constrain(speed_3,-50,50);
	motor_Data[1][3] = constrain(speed_4,-50,50);
}

void CodingDrone::Motor_Rotation(int speed_1, int speed_2, int speed_3, int speed_4) {
	motor_Data[2][0] = constrain(speed_1,-50,50);
	motor_Data[2][1] = constrain(speed_2,-50,50);
	motor_Data[2][2] = constrain(speed_3,-50,50);
	motor_Data[2][3] = constrain(speed_4,-50,50);
}

void CodingDrone::Motor_UpDown(int speed_1, int speed_2, int speed_3, int speed_4) {
	motor_Data[3][0] = constrain(speed_1,-50,50);
	motor_Data[3][1] = constrain(speed_2,-50,50);
	motor_Data[3][2] = constrain(speed_3,-50,50);
	motor_Data[3][3] = constrain(speed_4,-50,50);
}

void CodingDrone::Motor(int speed_1, int speed_2, int speed_3, int speed_4) {
	if(Drone_motor_Flag == 1){
		if(motor_addr[0] < 4) motor_Data2[motor_addr[0]] += constrain(speed_1,-50,50);
		if(motor_addr[1] < 4) motor_Data2[motor_addr[1]] += constrain(speed_2,-50,50);
		if(motor_addr[2] < 4) motor_Data2[motor_addr[2]] += constrain(speed_3,-50,50);
		if(motor_addr[3] < 4) motor_Data2[motor_addr[3]] += constrain(speed_4,-50,50);
	}
}

// 이륙 / 착륙 / 비상정지 명령처리
void CodingDrone::TakeOff(int height) {
	rcData[4] = 2000;
	take_off_height = height;
}

void CodingDrone::Landing() {
	rcData[4] = 1500;
}

void CodingDrone::EmergencyStop() {
	rcData[4] = 1000;
}



// 코딩드론은 저항으로 LED가 연결되어 있습니다
// 아래의 함수를 활용하면 편리하게 LED를 켜고 끌 수 있습니다.

void CodingDrone::LED_OFF(int Pin){  digitalWrite(Pin, LOW); }
void CodingDrone::LED_ON(int Pin){  digitalWrite(Pin, HIGH); }




//// CodingDrone_AutoMove Class Function

CodingDrone_AutoMove::CodingDrone_AutoMove() {
	Loop_Stop_Flag = false;
}


void CodingDrone_AutoMove::begin(){
	DroneMainFunction.begin();
	DroneMainFunction.MotorSetting(6,10,5,9);
	moveCurrentTime = millis();
	movePreviousTime = millis();
	DroneMainFunction.Auto_Move_Enable(0);
	
	Hold(2);
}

void CodingDrone_AutoMove::setTrim(int x, int y){
	DroneMainFunction.setTrim(x, y);
}

void CodingDrone_AutoMove::BackgroundProcess(){
	
	// 드론 배터리의 전압을 체크합니다.
	datadata = DroneMainFunction.readBattery(analogRead(A5));
	
	// 드론이 정상상태가 아닐 때 비상정지를 수행합니다.
	DroneMainFunction.FailSafe_NoneComm();
	
	//DroneMainFunction.readRemoteData();
	
	// 블루투스를 통해 최초로 통신명령이 들어오면 센서초기화를 수행합니다.
	DroneMainFunction.sensorCalibration();
	
	DroneMainFunction.AutoControl_None_Comm();
}

void CodingDrone_AutoMove::Start(){
	if(Loop_Stop_Flag){
		DroneMainFunction.Auto_Move_Enable(0);
	}
	else{
		DroneMainFunction.Auto_Move_Enable(1);
		
		movePreviousTime = millis();
		
		while (1) {
			moveCurrentTime = millis();
			
			Left_Right_Stop();
			Forward_Backward_Stop();
			Rotation_Stop();
			Up_Down_Stop();
			
			BackgroundProcess();
			
			if (DroneMainFunction.Drone_State_Flag == READY_TO_TAKEOFF) {
			  break;
			}
			else {
			  if (moveCurrentTime >= movePreviousTime + 10000) {
				  DroneMainFunction.Auto_Move_Enable(0);
				  break;
			  }
			}
		}
	}
	
}


void CodingDrone_AutoMove::End(){
	Loop_Stop_Flag = true;
	Hold(0.1);
}

void CodingDrone_AutoMove::TakeOff(){
	if(!Loop_Stop_Flag) DroneMainFunction.TakeOff(100);
	else DroneMainFunction.Landing();
	Hold(1);
}

void CodingDrone_AutoMove::Landing(){
	movePreviousTime = millis();
	DroneMainFunction.Landing();
	while (1) {
		moveCurrentTime = millis();
		
		Left_Right_Stop();
		Forward_Backward_Stop();
		Rotation_Stop();
		Up_Down_Stop();
		
		BackgroundProcess();
		
		if (DroneMainFunction.Drone_State_Flag == LANDING) {
		  if (moveCurrentTime >= movePreviousTime + 10000) {
			break;
		  }
		}
		else {
		  DroneMainFunction.Auto_Move_Enable(0);
		  break;
		}
	}
}

void CodingDrone_AutoMove::Hold(float sec){
	uint16_t millisec = abs(sec*1000);
	movePreviousTime = millis();
	
	while (1) {
		moveCurrentTime = millis();
		
		Left_Right_Stop();
		Forward_Backward_Stop();
		Rotation_Stop();
		Up_Down_Stop();
		
		BackgroundProcess();
		
		if (moveCurrentTime >= movePreviousTime + millisec) {
			break;
		}
	}
}

void CodingDrone_AutoMove::Left(float sec){
	if(!Loop_Stop_Flag) {
		uint16_t millisec = abs(sec*1000);
		movePreviousTime = millis();
		
		while (1) {
			moveCurrentTime = millis();
			
			Move_Left();
			Forward_Backward_Stop();
			Rotation_Stop();
			Up_Down_Stop();
			
			BackgroundProcess();
			
			if (moveCurrentTime >= movePreviousTime + millisec) {
				break;
			}
		}
	}
}

void CodingDrone_AutoMove::Right(float sec){
	if(!Loop_Stop_Flag) {
		uint16_t millisec = abs(sec*1000);
		movePreviousTime = millis();
		
		while (1) {
			moveCurrentTime = millis();
			
			Move_Right();
			Forward_Backward_Stop();
			Rotation_Stop();
			Up_Down_Stop();
			
			BackgroundProcess();
			
			if (moveCurrentTime >= movePreviousTime + millisec) {
				break;
			}
		}
	}
}

void CodingDrone_AutoMove::Forward(float sec){
	if(!Loop_Stop_Flag) {
		uint16_t millisec = abs(sec*1000);
		movePreviousTime = millis();
		
		while (1) {
			moveCurrentTime = millis();
			
			Left_Right_Stop();
			Move_Forward();
			Rotation_Stop();
			Up_Down_Stop();
			
			BackgroundProcess();
			
			if (moveCurrentTime >= movePreviousTime + millisec) {
				break;
			}
		}
	}
}

void CodingDrone_AutoMove::Backward(float sec){
	if(!Loop_Stop_Flag) {
		uint16_t millisec = abs(sec*1000);
		movePreviousTime = millis();
		
		while (1) {
			moveCurrentTime = millis();
			
			Left_Right_Stop();
			Move_Backward();
			Rotation_Stop();
			Up_Down_Stop();
			
			BackgroundProcess();
			
			if (moveCurrentTime >= movePreviousTime + millisec) {
				break;
			}
		}
	}
}

void CodingDrone_AutoMove::CW_Rotation(float sec){
	if(!Loop_Stop_Flag) {
		uint16_t millisec = abs(sec*1000);
		movePreviousTime = millis();
		
		while (1) {
			moveCurrentTime = millis();
			
			Left_Right_Stop();
			Forward_Backward_Stop();
			Move_CW_Rotation();
			Up_Down_Stop();
			
			BackgroundProcess();
			
			if (moveCurrentTime >= movePreviousTime + millisec) {
				break;
			}
		}
	}
}

void CodingDrone_AutoMove::CCW_Rotation(float sec){
	if(!Loop_Stop_Flag) {
		uint16_t millisec = abs(sec*1000);
		movePreviousTime = millis();
		
		while (1) {
			moveCurrentTime = millis();
			
			Left_Right_Stop();
			Forward_Backward_Stop();
			Move_CCW_Rotation();
			Up_Down_Stop();
			
			BackgroundProcess();
			
			if (moveCurrentTime >= movePreviousTime + millisec) {
				break;
			}
		}
	}
}

void CodingDrone_AutoMove::Up(float sec){
	if(!Loop_Stop_Flag) {
		uint16_t millisec = abs(sec*1000);
		movePreviousTime = millis();
		
		while (1) {
			moveCurrentTime = millis();
			
			Left_Right_Stop();
			Forward_Backward_Stop();
			Rotation_Stop();
			Move_Up();
			
			BackgroundProcess();
			
			if (moveCurrentTime >= movePreviousTime + millisec) {
				break;
			}
		}
	}
}

void CodingDrone_AutoMove::Down(float sec){
	if(!Loop_Stop_Flag) {
		uint16_t millisec = abs(sec*1000);
		movePreviousTime = millis();
		
		while (1) {
			moveCurrentTime = millis();
			
			Left_Right_Stop();
			Forward_Backward_Stop();
			Rotation_Stop();
			Move_Down();
			
			BackgroundProcess();
			
			if (moveCurrentTime >= movePreviousTime + millisec) {
				break;
			}
		}
	}
}



//좌측/우측 이동
void CodingDrone_AutoMove::Move_Left() {
	DroneMainFunction.Motor_LeftRight(-25,25,-25,25);
}

void CodingDrone_AutoMove::Move_Right() {
	DroneMainFunction.Motor_LeftRight(25,-25,25,-25);
}

void CodingDrone_AutoMove::Left_Right_Stop() {
	DroneMainFunction.Motor_LeftRight(0,0,0,0);
}

//전진/후진
void CodingDrone_AutoMove::Move_Forward() {
	DroneMainFunction.Motor_ForwardBackward(-25,-25,25,25);
}

void CodingDrone_AutoMove::Move_Backward() {
	DroneMainFunction.Motor_ForwardBackward(25,25,-25,-25);
}

void CodingDrone_AutoMove::Forward_Backward_Stop() {
	DroneMainFunction.Motor_ForwardBackward(0,0,0,0);
}

//시계방향/반시계방향 회전
void CodingDrone_AutoMove::Move_CW_Rotation() {
	DroneMainFunction.Motor_Rotation(25,-25,-25,25);
}

void CodingDrone_AutoMove::Move_CCW_Rotation() {
	DroneMainFunction.Motor_Rotation(-25,25,25,-25);
}

void CodingDrone_AutoMove::Rotation_Stop() {
	DroneMainFunction.Motor_Rotation(0,0,0,0);
}

//고도 상승/하강
void CodingDrone_AutoMove::Move_Up() {
	DroneMainFunction.Motor_UpDown(30,30,30,30);
}

void CodingDrone_AutoMove::Move_Down() {
	DroneMainFunction.Motor_UpDown(-25,-25,-25,-25);
}

void CodingDrone_AutoMove::Up_Down_Stop() {
	DroneMainFunction.Motor_UpDown(0,0,0,0);
}


