#include <avr/io.h>
#include "Arduino.h"
#include "BNC_CodingDrone.h"
#include "BNC_CodingDrone_Process.h"
#include "BNC_CodingDrone_Sensors.h"
#include "BNC_CodingDrone_def.h"
#include "BNC_CodingDrone_structs.h"
//#include <avr/pgmspace.h>

/*** HEADFREE : the copter can be controled by an absolute stick orientation, whatever the yaw orientation ***/
//#define HEADFREE

  // PITCH & ROLL only dynamic PID adjustemnt,  depending on throttle value (or collective.pitch value for heli)
  #define DYN_THR_PID_CHANNEL THROTTLE
  
// *************************
// DRONE TIME VARIABLE
// *************************


uint32_t currentTime = 0;
uint16_t previousTime = 0;
uint16_t cycleTime = 0;     // this is the number in micro second to achieve a full loop, it can differ a little and is taken into account in the PID loop
uint16_t cycleTime_TOTAL = 0;     // this is the number in micro second to achieve a full loop, it can differ a little and is taken into account in the PID loop
uint16_t rcTime  = 0;

// *************************
// DRONE SYSTEM VARIABLE
// *************************

uint8_t  vbatMin = VBATNOMINAL;  // lowest battery voltage in 0.1V steps
uint8_t  rcOptions[CHECKBOXITEMS];

int16_t  Idle_Throttle = 1100;
float  BaroPID = 0;
//int16_t  errorAltitudeI = 0;

int16_t rcCommand[4] = {1500, 1500, 1500, 1500};     // interval [1000;2000] for THROTTLE and [-500;+500] for ROLL/PITCH/YAW

// 통신변수

//uint8_t Serial_buffer[20] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
//uint8_t RC_buffer[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
//uint8_t rcSerial_ReadSeq = 0;
//int16_t rcSerial[8] = {1500, 1500, 1500, 1500, 1500, 1500, 1500, 1500};  // interval [1000;2000] - is rcData coming from MSP

uint8_t RC_buffer[6] = {0, 0, 0, 0, 0, 0};
uint8_t checksum = 0;

uint8_t Monitor_sequence = 0;
int16_t Alt_Throttle_Compensation = 0;
uint8_t Impact_Flag = false;
uint8_t FailSafe_Flag = false;

uint8_t Need_Calibration_Flag = true;

int8_t Drone_FastReady_Flag = false;

uint16_t Calibration_Delay_Count = 0;
uint8_t Reset_Att_Data_Flag = 1;
int8_t Drone_State_Flag_Backup = STATE_IDLE;

// *************************
// DRONE SYSTEM OBJECT
// *************************

System_Flag f;
analog_t analog;
global_conf_t global_conf;
conf_t conf;


// *************************
// MAIN LOOP VARIABLE
// *************************

int16_t rc;

int axisPID[3];

union {
  float fval;
  byte bval[4];
} floatAsBytes;

union {
  int ival;
  byte bval[2];
} intAsBytes;


///////////////////////////////////////////////////////

// *****************************
// Process Variables definition
// *****************************

int16_t initialThrottleHold;
volatile int16_t failsafeCnt = 0;

int16_t  magHold,headFreeModeHold; // [-180;+180]

uint8_t dynP8[2], dynD8[2];
uint32_t calibratedAccTime;
int16_t delta;
static uint8_t rcSticks;       // this hold sticks position for command combos  
static uint8_t rcDelayCommand; // this indicates the number of time (multiple of RC measurement at 50Hz) the sticks must be maintained to run or switch off motors
static int16_t lastGyro[2] = {0,0};
static int16_t errorAngleI[2] = {0,0};
  #if PID_CONTROLLER == 1
static int32_t errorGyroI_YAW;
static int16_t delta1[2],delta2[2];
static int32_t errorGyroI[2] = {0,0};
  #elif PID_CONTROLLER == 2
  static int16_t delta1[3],delta2[3];
  static int32_t errorGyroI[3] = {0,0,0};
  static int16_t lastError[3] = {0,0,0};
  int16_t deltaSum;
  int16_t AngleRateTmp, RateError;
  #endif
int16_t PTerm = 0,ITerm = 0,DTerm, PTermACC, ITermACC;
float PTerm_yaw = 0.0f,ITerm_yaw = 0.0f;
int16_t error,errorAngle;
int32_t prop = 0;  

uint8_t ARM_DELAY_FOR_CALIB = 100;

int16_t Throttle_TakeOff_Landing = 1100;

uint8_t Throttle_Control_for_TakeOff_Landing = 200;

int16_t lookupPitchRollRC[5];// lookup table for expo & RC rate PITCH+ROLL
uint16_t lookupThrottleRC[11];// lookup table for expo & mid THROTTLE

int32_t  AltHold = 0; // in cm
int32_t  AltVelocityHold = 0; // in cm
int32_t  AltHold_MAX = 250; // in cm
static uint8_t Alt_Throttle_Offset_Count = 0;
static int16_t Alt_Throttle_Offset = 0;
uint8_t Alt_Data_Process_Flag = 0;



static volatile uint8_t serialHeadRX[UART_NUMBER], serialTailRX[UART_NUMBER];
uint8_t serialBufferRX[RX_BUFFER_SIZE][UART_NUMBER];
static volatile uint8_t serialHeadTX[UART_NUMBER], serialTailTX[UART_NUMBER];
uint8_t serialBufferTX[TX_BUFFER_SIZE][UART_NUMBER];

uint8_t Serial_Parsing_Ready = 0;
uint8_t rcData_Ready_to_Read = 0;
uint8_t rcSerial_State = 0;
uint8_t rcSerial_Size = 0;
uint8_t send_data_count = 0; // 전송카운트


#if defined(THROTTLE_ANGLE_CORRECTION)
  int16_t throttleAngleCorrection = 0; // correction of throttle in lateral wind,
  int8_t  cosZ = 100;                  // cos(angleZ)*100
#else
  int8_t  cosZ = 512;                  // cos(angleZ)*100
#endif

uint8_t check_time_1 = 0;
uint8_t check_time_2 = 0; // 전송카운트


// *************************
// 시스템 동기화 함수
// *************************

void CodingDrone::DRONE_SYSTEM_PERIOD_LOOPTIME()
{
  while (1) {
    currentTime = micros();
    cycleTime = currentTime - previousTime;
    if (cycleTime >= LOOP_TIME) {
		break;
	}
  }
  // Time Variables Renewal.
  cycleTime_TOTAL = cycleTime;
  previousTime = currentTime;
}


void CodingDrone::DRONE_SYSTEM_PERIOD_20ms()
{
  // 현재시간 읽기
  currentTime = micros();
  if((int16_t)(currentTime - rcTime) > 0 ) // 50Hz - 초당 50번 주기로 실행
  {
    // RC Time Reset for 50Hz Control
	
    rcTime = currentTime + 20000;
	
	if(check_time_1 < 255) check_time_1++;
	if(check_time_2 < 255) check_time_2++;
	//return 1;
  }
  //else return 0;
}


// *************************
// 시스템 초기화 함수
// *************************

void CodingDrone::Initialize_BT()
{
  // 블루투스 통신용 시리얼포트 설정 및 블루투스 초기화
  // 블루투스가 처음 사용할 때는 9600으로 들어옵니다.
  // 드론에서는 속도를 올리기 위해서 115200으로 변경해야 하므로
  // 자동으로 변경하도록 이 내용이 추가되었습니다.
  delay(100);/*
  BT_SerialOpen(9600);
  delay(500);
  
  BT_SerialWrite('A');
  BT_SerialWrite('T');
  BT_SerialWrite('+');
  BT_SerialWrite('B');
  BT_SerialWrite('A');
  BT_SerialWrite('U');
  BT_SerialWrite('D');
  BT_SerialWrite('4');
  delay(500);
  BT_SerialWrite('A');
  BT_SerialWrite('T');
  BT_SerialWrite('+');
  BT_SerialWrite('R');
  BT_SerialWrite('E');
  BT_SerialWrite('S');
  BT_SerialWrite('E');
  BT_SerialWrite('T');
  delay(500);
  BT_SerialEnd();*/
  delay(500);
  BT_SerialOpen(115200);
  delay(10);
}


void CodingDrone::Initialize_InOutPins()
{
  // MOTOR PIN SETTING
  //pinMode(Pin_Motor_FL, OUTPUT);
  //pinMode(Pin_Motor_FR, OUTPUT);
  //pinMode(Pin_Motor_RL, OUTPUT);
  //pinMode(Pin_Motor_RR, OUTPUT);

  //digitalWrite(Pin_Motor_FL, LOW);
  //digitalWrite(Pin_Motor_FR, LOW);
  //digitalWrite(Pin_Motor_RL, LOW);
  //digitalWrite(Pin_Motor_RR, LOW);


  // LED PIN SETTING
  pinMode(Pin_LED_R, OUTPUT);
  pinMode(Pin_LED_G, OUTPUT);
  pinMode(Pin_LED_B, OUTPUT);
  pinMode(Pin_LED_Y, OUTPUT);
  
  digitalWrite(Pin_LED_R, LOW);
  digitalWrite(Pin_LED_G, LOW);
  digitalWrite(Pin_LED_B, LOW);
  digitalWrite(Pin_LED_Y, LOW);

  //시스템용 LED는 아두이노에서 일반적인 디지털 출력으로 동작시킬 수 없습니다.
  //시스템 상태에 따라 동작되도록 하기위해서 별도로 만들어진 함수가 있습니다.
  LEDPIN_PINMODE;
  
}


void CodingDrone::Bluetooth_Setting(String s)
{
	String Data = "AT+NAME";
	int i = 0;
  // 블루투스 통신용 시리얼포트 설정 및 블루투스 초기화
  // 블루투스가 처음 사용할 때는 9600으로 들어옵니다.
  // 드론에서는 속도를 올리기 위해서 115200으로 변경해야 하므로
  // 자동으로 변경하도록 이 내용이 추가되었습니다.
  
  delay(100);
  BT_SerialOpen(115200);
  Data += s;
  delay(100);
  BT_SerialWriteString("AT+RESET");
  
  delay(200);
  
  BT_SerialWriteString(Data);
    
  delay(500);
  BT_SerialWriteString("AT+RESET");
  delay(500);
  
  uint8_t buffffff = 0;
  while(BT_Available()){
	  buffffff = BT_SerialRead();
  }
}

float CodingDrone::readBattery(uint16_t VB)
{
  // query at most one multiplexed analog channel per MWii cycle

  static uint8_t ind = 0;
  static float vvec[VBAT_SMOOTH], vsum;
  uint16_t v = VB;
  float result_BAT = 0;

  vsum += v;
  vsum -= vvec[ind];
  vvec[ind++] = v;
  ind %= VBAT_SMOOTH;

  analog.vbat = vsum / 515 ; // result is Vbatt in 0.1V steps

  if (f.ARMED)  {
    if ( (analog.vbat > NO_VBAT) && (analog.vbat < vbatMin) ) vbatMin = analog.vbat;
  }
  
  result_BAT = analog.vbat/10;
  
  return result_BAT;
  
}


void CodingDrone::readRemoteData() {
	
  
  //static uint32_t nowTime = millis();
  //static uint16_t periodTime = millis()+20;

  //nowTime = millis();
  //if((int16_t)(nowTime - periodTime) > 0) {
	//periodTime = nowTime + 20;
  DRONE_SYSTEM_PERIOD_20ms();
  if(check_time_1 > 0) {
	check_time_1--;
	if (rcSerialCount < 80) rcSerialCount++;
	  else {
		f.COMM_ONLINE = false;
		f.COMM_ONLINE_BEFORE = false;
		Need_Calibration_Flag = true;
		
		cmd[ROLL] = 1500;
		cmd[PITCH] = 1500;
		cmd[YAW] = 1500;
		cmd[THROTTLE] = 1500;
		rcUnCorrectCount = 0;
		//cmd[AUX1] = 1500;
		/*
		uint8_t u = 0;
		uint8_t j = 0;
		
		for(u = 0;u < 4; u++){
			for(j = 0;j < 4; j++){
				motor_Data[u][j] = 0;
			}
		}*/
		
		if (f.ARMED == 0 && Drone_State_Flag != FAIL_SAFE) {
			Drone_State_Flag = COMM_READY;
			Drone_State_Flag_Backup = COMM_READY;
		}
	  }
	  
	  uint8_t data;
	  static uint8_t i;
	  
	  while (serialHeadRX[1] != serialTailRX[1])
	  {
		if(Drone_State_Flag == COMM_READY) rcSerialCount = 25;
		else rcSerialCount = 0;

		data = BT_SerialRead();
		//Serial.print(data);
		if (rcSerial_State == IDLE && data == '$'){
			rcSerial_State = HEADER_SIZE;
			
			if (f.COMM_ONLINE == 0){
				f.COMM_ONLINE++;
				Drone_State_Flag = COMM_READY;
				Drone_State_Flag_Backup = COMM_READY;
				cmd[ROLL] = 1500;
				cmd[PITCH] = 1500;
				cmd[YAW] = 1500;
				cmd[THROTTLE] = 1500;
				rcUnCorrectCount = 0;
				//cmd[AUX1] = 1500;
			}
		}
		else if (rcSerial_State == HEADER_SIZE)
		{
		  rcSerial_Size = data;//serialBufferRX[1][1];
		  if (rcSerial_Size > 6) rcSerial_State = IDLE;
		  else rcSerial_State = HEADER_CONTROL;
		  i = 0;
		  
		}
		else if (rcSerial_State == HEADER_CONTROL)
		{
			if (i < rcSerial_Size-1) {
				RC_buffer[i] = data;
				i++;
			}
			else{		
				RC_buffer[i] = data;
				checksum = 0xFF&(RC_buffer[0]+RC_buffer[1]+RC_buffer[2]+RC_buffer[3]+RC_buffer[4]+RC_buffer[5]);
				rcSerial_State = IDLE;
				if (checksum == 0) {
					rcUnCorrectCount = 0;
					for(i = 0;i < 5;i++) { cmd[i] = 1000 + (RC_buffer[i]*4); }
				}
				else{
					if(rcUnCorrectCount < 8) rcUnCorrectCount++;
					else {
						cmd[ROLL] = 1500;
						cmd[PITCH] = 1500;
						cmd[YAW] = 1500;
						cmd[THROTTLE] = 1500;
						//cmd[AUX1] = 1500;
					}
				}
				//Serial.println(rcUnCorrectCount);
			}
		}
	  }
	  
	// VBat information send
	if(f.COMM_ONLINE > 0){
		if (send_data_count == 6)
		{
		  //floatAsBytes.fval = rcData[THROTTLE];//analog.vbat;
		  floatAsBytes.fval = analog.vbat;
		  //byte *vbat_buffer = (byte *)&analog.vbat;
		  //BT_SerialWrite('$');
		  UDR1 = '$';
		  //UCSR1B &= ~(1 << UDRIE1);
		}
		else if (send_data_count == 5) {
			//BT_SerialWrite(floatAsBytes.bval[0]);
			UDR1 = floatAsBytes.bval[0];
			//UCSR1B &= ~(1 << UDRIE1);
		}
		else if (send_data_count == 4) {
			//BT_SerialWrite(floatAsBytes.bval[1]);
			UDR1 = floatAsBytes.bval[1];
			//UCSR1B &= ~(1 << UDRIE1);
		}
		else if (send_data_count == 3) {
			//BT_SerialWrite(floatAsBytes.bval[2]);
			UDR1 = floatAsBytes.bval[2];
			//UCSR1B &= ~(1 << UDRIE1);
		}
		else if (send_data_count == 2) {
			//BT_SerialWrite(floatAsBytes.bval[3]);
			UDR1 = floatAsBytes.bval[3];
			//UCSR1B &= ~(1 << UDRIE1);
		}
		else if (send_data_count == 1) {
			//BT_SerialWrite(Drone_State_Flag);
			UDR1 = Drone_State_Flag;
			//UCSR1B &= ~(1 << UDRIE1);
		}
		else if (send_data_count == 0) send_data_count = 7;
		send_data_count--;	
	}
	else send_data_count = 7;
	
  }
	
	  
  //}
}


void CodingDrone::Auto_Move_Enable(int gg) {

  if (gg) {
	f.COMM_ONLINE = 1;
  }
  else {
	f.COMM_ONLINE = false;
	f.COMM_ONLINE_BEFORE = false;
	Need_Calibration_Flag = true;
	
	if (f.ARMED == 0 && Drone_State_Flag != FAIL_SAFE) Drone_State_Flag = COMM_READY;
  }

}

void CodingDrone::sensorCalibration() {
	
  if(Drone_State_Flag == COMM_READY || Drone_State_Flag == FAIL_SAFE){
		if(Need_Calibration_Flag == false && calibratingA == 0 && calibratingG == 0) {
			if(Calibration_Delay_Count > 0) Calibration_Delay_Count--;
			else{
				if(f.SMALL_ANGLES_25) Calibration();
				else {
					Drone_State_Flag = READY_TO_TAKEOFF;
					Drone_State_Flag_Backup = READY_TO_TAKEOFF;
				}
			}
		}
	}
	
  if((f.COMM_ONLINE != 0 && f.COMM_ONLINE_BEFORE == 0) || sensorCalibrationSwitch == true){
	  Calibration();
	  sensorCalibrationSwitch = false;
	  f.COMM_ONLINE_BEFORE = f.COMM_ONLINE;
  }
  
}
/*
uint8_t CodingDrone::readCmd(uint8_t address ,int data){
	
	if(data > 1500){
		switch(address){
			case 0 : return Direction.Right;
				break;
			case 1 : return Direction.Forward;
				break;
			case 2 : return Direction.CW_Rotation;
				break;
			case 3 : return Direction.Up;
				break;
			case 4 : return System.TakeOff;
				break;
			default : break;
		}
	}
	else if(data < 1500){
		switch(address){
			case 0 : return Direction.Left;
				break;
			case 1 : return Direction.Backward;
				break;
			case 2 : return Direction.CCW_Rotation;
				break;
			case 3 : return Direction.Down;
				break;
			case 4 : return System.Emergency;
				break;
			default : break;
		}
	}
	else return Direction.None;
}
*/
/*
void CodingDrone::BAT_Data_to_App_Bluetooth(){
	
	static uint32_t nowTime = millis();
    static uint16_t periodTime = millis() + 50;

	nowTime = millis();
	if((int16_t)(nowTime - periodTime) > 0) {
		periodTime = nowTime + 50;
		// VBat information send
	}
	
	
}
*/

uint8_t CodingDrone::cmd_select(uint8_t c) {
	uint8_t databuffer = 0;
	
    switch(c){
		case 0 : databuffer = Direction.Left;
			break;
		case 1 : databuffer = Direction.Right;
			break;
		case 2 : databuffer = Direction.Forward;
			break;
		case 3 : databuffer = Direction.Backward;
			break;
		case 4 : databuffer = Direction.CCW_Rotation;
			break;
		case 5 : databuffer = Direction.CW_Rotation;
			break;
		case 6 : databuffer = Direction.Up;
			break;
		case 7 : databuffer = Direction.Down;
			break;
		case 8 : databuffer = System.TakeOff;
			break;
		case 9 : databuffer = System.Landing;
			break;
		default : break;
	}
	
	return databuffer;
}




void CodingDrone::init_Drone_Setting() {
  uint8_t i,y;
  int8_t tmp;
    go_disarm();
	initialThrottleHold = ALT_HOLD_THROTTLE_MIDPOINT;
    f.COMM_ONLINE = false;
	f.COMM_ONLINE_BEFORE = false;
	f.SLOW_START = 0;
	f.MOTOR_ON = 0;
	f.START_CALIBRATED_A = 0;
	f.START_CALIBRATED_G = 0;
	ARM_DELAY_FOR_CALIB = 100;
      // conf.pid[ROLL].P8     = 33;  conf.pid[ROLL].I8    = 30; conf.pid[ROLL].D8     = 23;
      // conf.pid[PITCH].P8    = 33; conf.pid[PITCH].I8    = 30; conf.pid[PITCH].D8    = 23;
      // conf.pid[PIDLEVEL].P8 = 90; conf.pid[PIDLEVEL].I8 = 10; conf.pid[PIDLEVEL].D8 = 100;
	  
	//conf.pid[ROLL].P8     = 18;  conf.pid[ROLL].I8    = 95; conf.pid[ROLL].D8     = 8;
    //conf.pid[PITCH].P8    = 18; conf.pid[PITCH].I8    = 95; conf.pid[PITCH].D8    = 8;
	conf.pid[ROLL].P8     = 25;  conf.pid[ROLL].I8    = 95; conf.pid[ROLL].D8     = 10;
    conf.pid[PITCH].P8    = 25; conf.pid[PITCH].I8    = 95; conf.pid[PITCH].D8    = 10;
	//conf.pid[ROLL].P8     = 22;  conf.pid[ROLL].I8    = 92; conf.pid[ROLL].D8     = 8;
    //conf.pid[PITCH].P8    = 22; conf.pid[PITCH].I8    = 92; conf.pid[PITCH].D8    = 8;
    
    conf.pid[PIDLEVEL].P8 = 150; conf.pid[PIDLEVEL].I8 = 10; conf.pid[PIDLEVEL].D8 = 100;
    
    //conf.pid[YAW].P8      = 120;  conf.pid[YAW].I8     = 0;  conf.pid[YAW].D8     = 8;
	conf.pid[YAW].P8      = 200;  conf.pid[YAW].I8     = 100;  conf.pid[YAW].D8     = 0;
    //conf.pid[PIDALT].P8   = 75; conf.pid[PIDALT].I8   = 15; conf.pid[PIDALT].D8   = 13; //Save 1.
    //conf.pid[PIDALT].P8   = 73; conf.pid[PIDALT].I8   = 16; conf.pid[PIDALT].D8   = 13; //Save 2.
    //conf.pid[PIDALT].P8   = 75; conf.pid[PIDALT].I8   = 15; conf.pid[PIDALT].D8   = 20; //Save 3.
    //conf.pid[PIDALT].P8   = 70; conf.pid[PIDALT].I8   = 15; conf.pid[PIDALT].D8   = 18; //Save 4.
    //conf.pid[PIDALT].P8   = 95; conf.pid[PIDALT].I8   = 20; conf.pid[PIDALT].D8   = 20; //Save 5.
    //conf.pid[PIDALT].P8   = 105; conf.pid[PIDALT].I8   = 16; conf.pid[PIDALT].D8   = 20; //Save 6.
    //conf.pid[PIDALT].P8   = 30; conf.pid[PIDALT].I8   = 21; conf.pid[PIDALT].D8   = 18; //Save 7.
	//conf.pid[PIDALT].P8   = 75; conf.pid[PIDALT].I8   = 21; conf.pid[PIDALT].D8   = 18; //Save 8.
	//conf.pid[PIDALT].P8   = 85; conf.pid[PIDALT].I8   = 15; conf.pid[PIDALT].D8   = 16; //Save 9.
	//conf.pid[PIDALT].P8   = 75; conf.pid[PIDALT].I8   = 25; conf.pid[PIDALT].D8   = 20; //Save 10.
	//conf.pid[PIDALT].P8   = 70; conf.pid[PIDALT].I8   = 28; conf.pid[PIDALT].D8   = 21; //Save 11.
	//conf.pid[PIDALT].P8   = 55; conf.pid[PIDALT].I8   = 35; conf.pid[PIDALT].D8   = 22; //Save 12.
	//conf.pid[PIDALT].P8   = 50; conf.pid[PIDALT].I8   = 33; conf.pid[PIDALT].D8   = 22; //Save 13.
	//conf.pid[PIDALT].P8   = 55; conf.pid[PIDALT].I8   = 35; conf.pid[PIDALT].D8   = 35; //Save 14. -> 2021 Final version.
	conf.pid[PIDALT].P8   = 55; conf.pid[PIDALT].I8   = 45; conf.pid[PIDALT].D8   = 37; //Save 15. -> 2022. 06. version.
	conf.pid[PIDVEL].P8   = 1; conf.pid[PIDVEL].I8   = 0; conf.pid[PIDVEL].D8   = 0;
	
    conf.pid[PIDMAG].P8   = 250;


    conf.rcRate8 = 20; conf.rcExpo8 = 0;
    conf.rollPitchRate = 0;
    conf.yawRate = 0;
    conf.dynThrPID = 0;
    conf.thrMid8 = 50; conf.thrExpo8 = 0;
    for(i=0;i<CHECKBOXITEMS;i++) {conf.activate[i] = 0;}
    conf.activate[BOXARM] = 4;  ////// BNC Default Setting
    conf.angleTrim[0] = ROLL_TRIM; conf.angleTrim[1] = PITCH_TRIM;
    conf.powerTrigger1 = 0;
  
  update_constants();
  
    // 500/128 = 3.90625    3.9062 * 3.9062 = 15.259   1526*100/128 = 1192
  for(i=0;i<5;i++) {
    lookupPitchRollRC[i] = (1526+conf.rcExpo8*(i*i-15))*i*(int32_t)conf.rcRate8/1192;
  }
  for(i=0;i<11;i++) {
    tmp = 10*i-conf.thrMid8;
    y = conf.thrMid8;
    if (tmp>0) y = 100-y;
    lookupThrottleRC[i] = 100*conf.thrMid8 + tmp*( (int32_t)conf.thrExpo8*(tmp*tmp)/((uint16_t)y*y)+100-conf.thrExpo8 );       // [0;10000]
    lookupThrottleRC[i] = conf.minthrottle + (uint32_t)((uint16_t)(MAXTHROTTLE-conf.minthrottle))* lookupThrottleRC[i]/10000;  // [0;10000] -> [conf.minthrottle;MAXTHROTTLE]
  }
  
  delay(10);
  
  //loop()함수에서 주기적인 동작이 이루어지도록 마이크로 초 단위의 시간을 저장합니다.
  previousTime = micros();
  currentTime = micros();
  rcTime = currentTime + 20000;
  
}

void update_constants() { 
  #if defined (FAILSAFE)
    conf.failsafe_throttle = FAILSAFE_THROTTLE;
  #endif
  conf.vbatscale = VBATSCALE;
  conf.vbatlevel_warn1 = VBATLEVEL_WARN1;
  conf.vbatlevel_warn2 = VBATLEVEL_WARN2;
  conf.vbatlevel_crit = VBATLEVEL_CRIT;
  
  #ifdef MMGYRO
    conf.mmgyro = MMGYRO;
  #endif
  #if defined(ARMEDTIMEWARNING)
    conf.armedtimewarning = ARMEDTIMEWARNING;
  #endif
  conf.minthrottle = MINTHROTTLE;
}

void Calibration()
{
	uint8_t for_i = 0;
	
	calibratingA=128;
	calibratingG=128;
	calibratingB=5;
	FailSafe_Flag = 0;
	att.heading = 0;
	Need_Calibration_Flag = false;
	
	Calibration_Delay_Count = 1000;
	Reset_Att_Data_Flag = true;
	//Drone_State_Flag = READY_TO_TAKEOFF;
	/*
	if(calibratingA == 0 && calibratingG == 0) 
	{
		f.ACC_CALIBRATED = 1;
		f.SLOW_START = 0;
		f.MOTOR_ON = 0;
		f.START_CALIBRATED_A = 0;
		f.START_CALIBRATED_G = 0;
		ARM_DELAY_FOR_CALIB = 100;
	}
	*/
	for(for_i = 0;for_i < 3;for_i++){
		imu.accSmooth[for_i] = 0;
		imu.gyroData[for_i] = 0;
		imu.magADC[for_i] = 0;
		imu.gyroADC[for_i] = 0;
		imu.accADC[for_i] = 0;
	}
}

void CodingDrone::writeMoveCmd_Pitch(int data){
	rcData[PITCH] = 1000 + (int16_t)(data << 2);
}
void CodingDrone::writeMoveCmd_Roll(int data){
	rcData[ROLL] = 1000 + (int16_t)(data << 2);
}
void CodingDrone::writeMoveCmd_Yaw(int data){
	rcData[YAW] = 1000 + (int16_t)(data << 2);
}
void CodingDrone::writeMoveCmd_Throttle(int data){
	rcData[THROTTLE] = 1000 + (int16_t)(data << 2);
}
void CodingDrone::writeMoveCmd_System(int data){
	rcData[AUX1] = 1000 + (int16_t)(data << 2);
}

void CodingDrone::FailSafe()
{
	
	//static uint32_t nowTime = micros();
    //static uint16_t periodTime  = micros()+20000;

	//nowTime = micros();
	//if((int16_t)(nowTime - periodTime) > 0) {
		
		//periodTime = nowTime + 20000;
		// Failsafe routine
		if(Drone_State_Flag == FAIL_SAFE && rcData[AUX1] == 1000) {
			rcData[AUX1] = 1500;
			Drone_State_Flag = COMM_READY;
			Drone_State_Flag_Backup = COMM_READY;
		}
		
		if(f.COMM_ONLINE == false) rcOptions[BOXARM] = 0;	
		
		  if(f.ARMED)
		  {
			  if(FailSafe_Flag) {
				rcOptions[BOXARM] = 0;
				f.ARMED = 0;
				Drone_State_Flag = FAIL_SAFE;
				Drone_State_Flag_Backup = FAIL_SAFE;
				Need_Calibration_Flag = true;
				FailSafe_Flag = 0;
				//Drone_FastReady_Flag = true;
			  }
			  if(f.COMM_ONLINE == false) {
				rcOptions[BOXARM] = 0;
				f.ARMED = 0;
				Drone_State_Flag = COMM_READY;
				Drone_State_Flag_Backup = COMM_READY;
				Need_Calibration_Flag = true;
				//Drone_FastReady_Flag = true;
			  }
		  }
		// end of failsafe routine
		
	//}
	
}


void CodingDrone::FailSafe_NoneComm()
{
		// Failsafe routine
		if(Drone_State_Flag == FAIL_SAFE) rcData[AUX1] = 1000;
		
		if(f.COMM_ONLINE == false) rcOptions[BOXARM] = 0;	
		
		  if(f.ARMED)
		  {
			  if(FailSafe_Flag) {
				rcOptions[BOXARM] = 0;
				f.ARMED = 0;
				Drone_State_Flag = FAIL_SAFE;
				Need_Calibration_Flag = true;
				//Drone_FastReady_Flag = true;
			  }
		  }
		// end of failsafe routine
		
		
}

void CodingDrone::Attitude_PID()
{
  //**** PITCH & ROLL & YAW PID ****
  uint8_t axis;	
  
  
  #if PID_CONTROLLER == 1 // evolved oldschool
  
  prop = min(max(abs(rcCommand[PITCH]),abs(rcCommand[ROLL])),512);
  // PITCH & ROLL
  for(axis=0;axis<2;axis++) {
    rc = rcCommand[axis]<<1;
    error = rc - imu.gyroData[axis];
    errorGyroI[axis]  = constrain(errorGyroI[axis]+error,-16000,+16000);       // WindUp   16 bits is ok here
    if (abs(imu.gyroData[axis])>640) errorGyroI[axis] = 0;

    ITerm = (errorGyroI[axis]>>7)*conf.pid[axis].I8>>6;                        // 16 bits is ok here 16000/125 = 128 ; 128*250 = 32000

    PTerm = mul(rc,conf.pid[axis].P8)>>6;
    
    // 50 degrees max inclination
    errorAngle         = constrain(rc,-500,+500) - att.angle[axis] + conf.angleTrim[axis]; //16 bits is ok here
    errorAngleI[axis]  = constrain(errorAngleI[axis]+errorAngle,-10000,+10000);                                                // WindUp     //16 bits is ok here

    PTermACC           = mul(errorAngle,conf.pid[PIDLEVEL].P8)>>7; // 32 bits is needed for calculation: errorAngle*P8 could exceed 32768   16 bits is ok for result

    int16_t limit      = conf.pid[PIDLEVEL].D8*5;
    PTermACC           = constrain(PTermACC,-limit,+limit);

    ITermACC           = mul(errorAngleI[axis],conf.pid[PIDLEVEL].I8)>>12;   // 32 bits is needed for calculation:10000*I8 could exceed 32768   16 bits is ok for result

    ITerm              = ITermACC + ((ITerm-ITermACC)*prop>>9);
    PTerm              = PTermACC + ((PTerm-PTermACC)*prop>>9);
    

    PTerm -= mul(imu.gyroData[axis],dynP8[axis])>>6; // 32 bits is needed for calculation   

    delta          = imu.gyroData[axis] - lastGyro[axis];  // 16 bits is ok here, the dif between 2 consecutive gyro reads is limited to 800
    lastGyro[axis] = imu.gyroData[axis];
    DTerm          = delta1[axis]+delta2[axis]+delta;
    delta2[axis]   = delta1[axis];
    delta1[axis]   = delta;
 
    DTerm = mul(DTerm,dynD8[axis])>>5;        // 32 bits is needed for calculation

    axisPID[axis] =  PTerm + ITerm - DTerm;
  }

  //YAW
  #define GYRO_P_MAX 256
  #define GYRO_I_MAX 250
  
  
  if (abs(rcData[YAW]-1500) < 2 ) {
    int32_t dif = att.heading - magHold;
    if (dif <= - 1800) dif += 3600;
    if (dif >= + 1800) dif -= 3600;
    rcCommand[YAW] -= constrain(dif*conf.pid[PIDMAG].P8/200,-500,500);  //Always correct maghold in GPS mode
  } else magHold = att.heading;
    
  //rc = mul(rcCommand[YAW] , (2*conf.yawRate + 30))  >> 5;
  rc = rcCommand[YAW];
  //error = rc - imu.gyroData[YAW] + imu.gyroADC_offset[YAW];
  //error = rc - imu.gyroData[YAW];
  if (abs(imu.gyroData[YAW]) > 3) error = rc - imu.gyroData[YAW];
  else error = rc;
  
  if(f.ARMED) errorGyroI_YAW += mul(error,conf.pid[YAW].I8);
  else errorGyroI_YAW = 0;
  errorGyroI_YAW  = constrain(errorGyroI_YAW, 2-((int32_t)1<<15), -2+((int32_t)1<<15));
  
  if (abs(rc) > 50) errorGyroI_YAW = 0;
  
  PTerm_yaw = mul(error,conf.pid[YAW].P8)/100;
  
  //ITerm_yaw = (float)errorGyroI_YAW/10000;//constrain((int16_t)(errorGyroI_YAW>>13),-GYRO_I_MAX,+GYRO_I_MAX);
  ITerm_yaw = constrain((float)errorGyroI_YAW/10000,-GYRO_I_MAX,+GYRO_I_MAX);
  
  axisPID[YAW] = PTerm_yaw + ITerm_yaw;
  
  #elif PID_CONTROLLER == 2 // alexK
  
  #define GYRO_I_MAX 256
  #define ACC_I_MAX 256
  prop = min(max(abs(rcCommand[PITCH]),abs(rcCommand[ROLL])),500); // range [0;500]
  
  for(axis=0;axis<3;axis++) {
	
	errorAngle = constrain((rcCommand[axis]<<1),-500,+500) - att.angle[axis] + conf.angleTrim[axis]; //16 bits is ok here
	
	if (axis == 2) {//YAW is always gyro-controlled (MAG correction is applied to rcCommand)
      AngleRateTmp = (((int32_t) (conf.yawRate + 27) * rcCommand[2]) >> 5);
    } else {
		//control is GYRO based (ACRO and HORIZON - direct sticks control is applied to rate PID
		//AngleRateTmp = ((int32_t) (conf.rollPitchRate + 27) * rcCommand[axis]) >> 4;
		//AngleRateTmp += ((int32_t) errorAngle * conf.pid[PIDLEVEL].I8)>>8;
		
        //it's the ANGLE mode - control is angle based, so control loop is needed
        AngleRateTmp = ((int32_t) errorAngle * conf.pid[PIDLEVEL].P8)>>4;
    }
	
    RateError = AngleRateTmp  - imu.gyroData[axis];
    //-----calculate P component
    PTerm = ((int32_t) RateError * conf.pid[axis].P8)>>7;
	
    //-----calculate I component
    errorGyroI[axis]  += (((int32_t) RateError * cycleTime)>>11) * conf.pid[axis].I8;
    errorGyroI[axis]  = constrain(errorGyroI[axis], (int32_t) -GYRO_I_MAX<<13, (int32_t) +GYRO_I_MAX<<13);
    ITerm = errorGyroI[axis]>>13;
	
    //-----calculate D-term
    delta          = RateError - lastError[axis];  // 16 bits is ok here, the dif between 2 consecutive gyro reads is limited to 800
    lastError[axis] = RateError;
	delta = ((int32_t) delta * ((uint16_t)0xFFFF / (cycleTime>>4)))>>6;
    //add moving average here to reduce noise
    deltaSum       = delta1[axis]+delta2[axis]+delta;
    delta2[axis]   = delta1[axis];
    delta1[axis]   = delta;

    //DTerm = (deltaSum*conf.pid[axis].D8)>>8;
    //Solve overflow in calculation above...
    DTerm = ((int32_t)deltaSum*conf.pid[axis].D8)>>8;
    //-----calculate total PID output
    axisPID[axis] =  PTerm + ITerm + DTerm;
  }

  #endif
  
}


void CodingDrone::go_arm() {

  if(calibratingA == 0 && calibratingG == 0 && calibratingB == 0 && Drone_State_Flag == READY_TO_TAKEOFF) {
    if(!f.ARMED) { // arm now!
	  f.MOTOR_ON = 1;
      f.ARMED = 1;
	  Drone_State_Flag = IN_FLIGHT;
	  Drone_State_Flag_Backup = IN_FLIGHT;
	  FailSafe_Flag = 0;
	  Impact_Flag = 0;
	  AltHold = alt.EstAlt+take_off_height+50; // for Take-Off
	  att.Yaw_Hold = att.heading;
	  //AltHold = alt.EstAlt; // for Take-Off
	  f.SLOW_START = 1;
      #if defined(HEADFREE)
        headFreeModeHold = att.heading;
      #endif
      magHold = att.heading;
      if (analog.vbat > NO_VBAT) vbatMin = analog.vbat;
	  //BT_SerialWrite('@');
      //calibratingB = 10; // calibrate baro to new ground level (10 * 25 ms = ~250 ms non blocking)
      
    }
  } 
}

void CodingDrone::go_disarm() {
  if (f.ARMED) {
	Drone_State_Flag = LANDING;
	Drone_State_Flag_Backup = LANDING;
	//  f.ARMED = 0;
	if(Impact_Flag) {
		f.ARMED = 0;
		Drone_State_Flag = COMM_READY;
		Drone_State_Flag_Backup = COMM_READY;
		Need_Calibration_Flag = true;
		//Drone_FastReady_Flag = true;
	}
	//if(att.Angle_Feedback < 490) f.ARMED = 0;
	else if(rcData[AUX1] < 1250) {
		f.ARMED = 0;
		Drone_State_Flag = COMM_READY;
		Drone_State_Flag_Backup = COMM_READY;
		Need_Calibration_Flag = true;
		//Drone_FastReady_Flag = true;
	}
  }
}

void CodingDrone::ALTITUDE_PROCESS()
{
    static uint8_t isAltHoldChanged = 0;
    int16_t AltHoldCorr = 0;
    int16_t UpDown_Throttle = 0;
	
	if (Alt_Data_Process_Flag > 0) { 
	  if (getEstimatedAltitude() != 0) 
	  {
	    Alt_Data_Process_Flag = 0;
		//IF Throttle not ignored then allow change altitude with the stick....
		
		AltHoldCorr = rcData[THROTTLE] - initialThrottleHold;
		if ( abs(AltHoldCorr) > 0) { 
		    AltHold = alt.EstAlt + (AltHoldCorr/18);
		    UpDown_Throttle = (AltHoldCorr/18);
	 /*
		  if(AltHoldCorr < 0) {
		      AltHold = alt.EstAlt - 5 + (AltHoldCorr/8);
			  UpDown_Throttle = AltHoldCorr/8;
		  }
		  else 
		  {
		      AltHold = alt.EstAlt + (5 + AltHoldCorr/10);
		      UpDown_Throttle = (AltHoldCorr/10);
		  }*/
		}
		else
		{
			UpDown_Throttle = 0;
		}
		
		if(rcOptions[BOXARM] == 0 && f.ARMED) 
		{
			AltHold = alt.EstAlt - 20;
			UpDown_Throttle = -20;
		}
		/*
		if ( (abs(rcCommand[THROTTLE]-initialThrottleHold)>ALT_HOLD_THROTTLE_NEUTRAL_ZONE)) {
		  // Slowly increase/decrease AltHold proportional to stick movement ( +100 throttle gives ~ +50 cm in 1 second with cycle time about 3-4ms)
		  AltHoldCorr+= rcCommand[THROTTLE] - initialThrottleHold;
		  if(abs(AltHoldCorr) > 512) {
			AltHold += AltHoldCorr/512;
			AltHoldCorr %= 512;
		  }
		  isAltHoldChanged = 1;
		} else if (isAltHoldChanged) {
		  AltHold = alt.EstAlt;
		  isAltHoldChanged = 0;
		}*/
	
		if(Drone_State_Flag == COMM_READY || Drone_State_Flag == READY_TO_TAKEOFF) BaroPID = 0;
		
		//rcCommand[THROTTLE] = 1500 + BaroPID*5 + UpDown_Throttle;
		rcCommand[THROTTLE] = 1500 + BaroPID*6 + UpDown_Throttle;
		//rcCommand[THROTTLE] = initialThrottleHold + BaroPID*5;
		
		#if defined(THROTTLE_ANGLE_CORRECTION)
		rcCommand[THROTTLE] += throttleAngleCorrection;
		#endif
	
		if(rcCommand[THROTTLE] > 1900) rcCommand[THROTTLE] = 1900; 
		if(rcCommand[THROTTLE] < 1250) rcCommand[THROTTLE] = 1250;
	  }
	}
}

uint8_t CodingDrone::Motor_select(int c) {
    if(c == 6) return 0;
	else if(c == 10) return 1;
	else if(c == 5) return 2;
	else if(c == 9) return 3;
	else return 4;
}

void CodingDrone::Drone_analogWrite(int pin, int pwm) {
	uint8_t addr = Motor_select(pin);
	if(Drone_motor_Flag && (addr < 4)) motor_Data2[addr] += constrain(pwm,-50,50);
	else analogWrite(pin,pwm);
}

void CodingDrone::readSensorData(){
	
  // 블루투스를 통해 들어오는 통신데이터를 읽습니다.
  //readRemoteData();
  //Remote_Read_to_Control_Data();
  // 드론이 일정 주기에 맞춰서 동작하도록 동기화 함수를 사용합니다.  
  // 기본값으로 2.7ms 주기로 동기화가 진행됩니다.
		
		uint16_t tmp,tmp2;
		uint8_t axis,prop1,prop2;
		//uint16_t auxState = 0;
		uint8_t i = 0;
		uint8_t j = 0;	
		
		
		//Left Right Control
		//rcData[ROLL] = 1500;
		rcData[ROLL] = constrain(1500 - (motor_Data2[0] - motor_Data2[1] + motor_Data2[2] - motor_Data2[3])*5/4,1100,1900);
		
		//Forward Backward Control
		//rcData[PITCH] = 1500;
		rcData[PITCH] = constrain(1500 - (motor_Data2[2] - motor_Data2[0] + motor_Data2[3] - motor_Data2[1])*5/4,1100,1900);
		
		//Rotation Control
		//rcData[YAW] = 1500;
		rcData[YAW] = constrain(1500 + (motor_Data2[1] - motor_Data2[0] - motor_Data2[3] + motor_Data2[2])*5/4,1100,1900);
		
		//Up Down Control
		//rcData[THROTTLE] = 1500;
		rcData[THROTTLE] = constrain(1500 + (motor_Data2[0] + motor_Data2[1] + motor_Data2[2] + motor_Data2[3])*5/4,1100,1900);
		
		motor_Data2[0] = 0;
		motor_Data2[1] = 0;
		motor_Data2[2] = 0;
		motor_Data2[3] = 0;
		
		if(!f.COMM_ONLINE || Drone_State_Flag == COMM_READY || Drone_State_Flag == FAIL_SAFE) rcData[AUX1] = 1500;


		  prop2 = 128; // prop2 was 100, is 128 now
		  if (rcData[DYN_THR_PID_CHANNEL]>1500) { // breakpoint is fix: 1500
			prop2 -=  ((uint16_t)conf.dynThrPID*(rcData[DYN_THR_PID_CHANNEL]-1500)>>9); //  /512 instead of /500
		  }
		  
		for(axis=0;axis<3;axis++) {
			tmp = min(abs(rcData[axis]-MIDRC),500);
			#if defined(DEADBAND)
			  if (tmp>DEADBAND) { tmp -= DEADBAND; }
			  else { tmp=0; }
			#endif
			if(axis!=2) { //ROLL & PITCH
			  tmp2 = tmp>>7; // 500/128 = 3.9  => range [0;3]
			  rcCommand[axis] = lookupPitchRollRC[tmp2] + ((tmp-(tmp2<<7)) * (lookupPitchRollRC[tmp2+1]-lookupPitchRollRC[tmp2])>>7);
			  prop1 = 128-((uint16_t)conf.rollPitchRate*tmp>>9); // prop1 was 100, is 128 now -- and /512 instead of /500
			  prop1 = (uint16_t)prop1*prop2>>7; // prop1: max is 128   prop2: max is 128   result prop1: max is 128
			  dynP8[axis] = (uint16_t)conf.pid[axis].P8*prop1>>7; // was /100, is /128 now
			  dynD8[axis] = (uint16_t)conf.pid[axis].D8*prop1>>7; // was /100, is /128 now
			} else {      // YAW
			  rcCommand[axis] = tmp;
			}
			
			if (rcData[axis]<MIDRC) rcCommand[axis] = -rcCommand[axis];
		}
		
		tmp = constrain(rcData[THROTTLE],MINCHECK,2000);
		tmp = (uint32_t)(tmp-MINCHECK)*2559/(2000-MINCHECK); // [MINCHECK;2000] -> [0;2559]
		tmp2 = tmp/256; // range [0;9]
		//rcCommand[THROTTLE] = lookupThrottleRC[tmp2] + (tmp-tmp2*256) * (lookupThrottleRC[tmp2+1]-lookupThrottleRC[tmp2]) / 256; // [0;2559] -> expo -> [conf.minthrottle;MAXTHROTTLE]
		//Serial.println(rcCommand[0]);
		
		
		
		if(rcData[AUX1] > 1750) rcOptions[0] = true;
		else rcOptions[0] = false;
		
		
		// perform actions
		  //errorGyroI[ROLL] = 0; errorGyroI[PITCH] = 0;
		  //errorGyroI_YAW = 0;
		  //errorAngleI[ROLL] = 0; errorAngleI[PITCH] = 0;
		  
		  if (conf.activate[BOXARM] > 0) {             // Arming/Disarming via ARM BOX
			
			if ( rcOptions[BOXARM]) {
				errorAngleI[ROLL] = 0; errorAngleI[PITCH] = 0;
				#if PID_CONTROLLER == 1 // evolved oldschool
				errorGyroI[ROLL] = 0; errorGyroI[PITCH] = 0;
				errorGyroI_YAW = 0;
				#elif PID_CONTROLLER == 2 // alexK
				errorGyroI[ROLL] = 0; errorGyroI[PITCH] = 0; errorGyroI[YAW] = 0;
				#endif
				go_arm(); 
			}
			else if (rcOptions[BOXARM] == 0 && f.ARMED) {
				go_disarm();
			}
			else {
				f.SLOW_START = 0;
				f.MOTOR_ON = 0;
				f.START_CALIBRATED_A = 0;
				f.START_CALIBRATED_G = 0;
				ARM_DELAY_FOR_CALIB = 20;
			}
		  }
		
		if (f.ARMED == 0) f.OK_TO_ARM = 1;
		
		System_LED_Control();
	
	
  DRONE_SYSTEM_PERIOD_LOOPTIME();
  
  // 드론의 현재 자세를 알기 위해 MPU6050 센서값을 읽고 각도와 방향을 확인합니다.
  Attitude_by_MPU6050();
  
  Attitude_PID();
  
  if (Read_AirPressure() == 2) {
	   if(Alt_Data_Process_Flag < 255 ) Alt_Data_Process_Flag++;
	   //Serial.println(alt.EstAlt);
	   //Serial.println(baroPressureSum);
	}
	ALTITUDE_PROCESS();
 

  // 고도와 자세제어를 위한 명령을 모터 회전값으로 계산하여 반영합니다.
  // PID_to_DRONE_MOTOR_SPEED();
  
  // 모터 회전값을 적용합니다.
  // DRONE_MOTOR_CONTROL();
}


void CodingDrone::Auto_PID_Control(int Roll_Data, int Pitch_Data, int Yaw_Data, int Alt_Data) {
	
  int16_t maxMotor;
  uint8_t i;
  
  //Attitude_PID();
  //ALTITUDE_PROCESS();
  //if(ALTITUDE_Data() == Alt_Data) ALTITUDE_PROCESS();
  
  
  //Drone_Motor_1 = 0;
  //Drone_Motor_2 = 0;
  //Drone_Motor_3 = 0;
  //Drone_Motor_4 = 0;
  
  /**************** main Mix Table ******************/
  // Apply PID Result Data -> QuadCopter Control Data
  // DRONE_MOTOR_PIDMIX(ROLL_DIRECTION, PITCH_DIRECTION, YAW_DIRECTION)
  motor[0] = DRONE_MOTOR_PIDMIX(-1, +1, -1); //FRONT_L
  motor[1] = DRONE_MOTOR_PIDMIX(+1, +1, +1); //FRONT_R
  motor[2] = DRONE_MOTOR_PIDMIX(-1, -1, +1); //REAR_L
  motor[3] = DRONE_MOTOR_PIDMIX(+1, -1, -1); //REAR_R
  
  /****************                normalize the Motors values                ******************/
  maxMotor=motor[0];
  
  for(i=1; i< NUMBER_MOTOR; i++)
	if (motor[i]>maxMotor) maxMotor=motor[i];
  
  for(i=0; i< NUMBER_MOTOR; i++) {
	if (maxMotor > MAXTHROTTLE) // this is a way to still have good gyro corrections if at least one motor reaches its max.
	  motor[i] -= maxMotor - MAXTHROTTLE;
	
	motor[i] = constrain(motor[i], conf.minthrottle, MAXTHROTTLE);
	if (!f.ARMED)
	{
		if(!f.MOTOR_ON) motor[i] = MINCOMMAND;
		else motor[i] = STANBYCOMMAND;
	}
  }
  
  if((Sensor_Data(0) == Roll_Data) && (Sensor_Data(1) == Pitch_Data) && (Sensor_Data(2) == Yaw_Data) && (Sensor_Data(3) == Alt_Data)) {
	//Drone_Motor_1 = map(motor[0], 1000, 2000, 0, 255);
    //Drone_Motor_2 = map(motor[1], 1000, 2000, 0, 255);
    //Drone_Motor_3 = map(motor[2], 1000, 2000, 0, 255);
    //Drone_Motor_4 = map(motor[3], 1000, 2000, 0, 255);
	  analogWrite(Pin_Motor_FL,map(motor[0], 1000, 2000, 0, 255));
	  analogWrite(Pin_Motor_FR,map(motor[1], 1000, 2000, 0, 255));
	  analogWrite(Pin_Motor_RL,map(motor[2], 1000, 2000, 0, 255));
	  analogWrite(Pin_Motor_RR,map(motor[3], 1000, 2000, 0, 255));
  }
  else {
	  //Drone_Motor_1 = 0;
	  //Drone_Motor_2 = 0;
	  //Drone_Motor_3 = 0;
	  //Drone_Motor_4 = 0;
	  analogWrite(Pin_Motor_FL,0);
	  analogWrite(Pin_Motor_FR,0);
	  analogWrite(Pin_Motor_RL,0);
	  analogWrite(Pin_Motor_RR,0);
  }
}


void CodingDrone::AutoControl_None_Comm(){
	
  int16_t maxMotor;
  uint8_t i;
	
	
  // 블루투스를 통해 들어오는 통신데이터를 읽습니다.

  // 드론이 일정 주기에 맞춰서 동작하도록 동기화 함수를 사용합니다.  
  // 기본값으로 2.7ms 주기로 동기화가 진행됩니다.
  DRONE_SYSTEM_PERIOD_LOOPTIME();
  
  // 드론의 현재 자세를 알기 위해 MPU6050 센서값을 읽고 각도와 방향을 확인합니다.
  Attitude_by_MPU6050();
	
  // 조종기 명령과 드론의 현재 자세를 합쳐서 드론의 자세를 제어합니다.
  Attitude_PID();
  
  
  ALTITUDE_PROCESS();

  // 고도와 자세제어를 위한 명령을 모터 회전값으로 계산하여 반영합니다.
  //PID_to_DRONE_MOTOR_SPEED();
  
  // 모터 회전값을 적용합니다.
  //DRONE_MOTOR_CONTROL();
  
  /**************** main Mix Table ******************/
  // Apply PID Result Data -> QuadCopter Control Data
  // DRONE_MOTOR_PIDMIX(ROLL_DIRECTION, PITCH_DIRECTION, YAW_DIRECTION)
  motor[0] = DRONE_MOTOR_PIDMIX(-1, +1, -1); //FRONT_L
  motor[1] = DRONE_MOTOR_PIDMIX(+1, +1, +1); //FRONT_R
  motor[2] = DRONE_MOTOR_PIDMIX(-1, -1, +1); //REAR_L
  motor[3] = DRONE_MOTOR_PIDMIX(+1, -1, -1); //REAR_R
  
  
  /****************                normalize the Motors values                ******************/
  maxMotor=motor[0];
  
  for(i=1; i< NUMBER_MOTOR; i++)
	if (motor[i]>maxMotor) maxMotor=motor[i];
  
  for(i=0; i< NUMBER_MOTOR; i++) {
	if (maxMotor > MAXTHROTTLE) // this is a way to still have good gyro corrections if at least one motor reaches its max.
	  motor[i] -= maxMotor - MAXTHROTTLE;
	
	motor[i] = constrain(motor[i], conf.minthrottle, MAXTHROTTLE);
	if (!f.ARMED)
	{
		if(!f.MOTOR_ON) motor[i] = MINCOMMAND;
		else motor[i] = STANBYCOMMAND;
	}
  }
  
  analogWrite(Pin_Motor_FL,map(motor[0], 1000, 2000, 0, 255));
  analogWrite(Pin_Motor_FR,map(motor[1], 1000, 2000, 0, 255));
  analogWrite(Pin_Motor_RL,map(motor[2], 1000, 2000, 0, 255));
  analogWrite(Pin_Motor_RR,map(motor[3], 1000, 2000, 0, 255));
}



void HEADFREE_PROCESS()
{
    float radDiff = (att.heading - headFreeModeHold) * 0.0174533f; // where PI/180 ~= 0.0174533
    float cosDiff = cos(radDiff);
    float sinDiff = sin(radDiff);
    int16_t rcCommand_PITCH = rcCommand[PITCH]*cosDiff + rcCommand[ROLL]*sinDiff;
    rcCommand[ROLL] =  rcCommand[ROLL]*cosDiff - rcCommand[PITCH]*sinDiff; 
    rcCommand[PITCH] = rcCommand_PITCH;  
}

void CodingDrone::System_LED_Control()
{
  // LEDPIN_ON : 시스템 LED 켜기 /  LEDPIN_OFF : 시스템 LED 끄기 /  LEDPIN_TOGGLE : 시스템 LED 상태 바꾸기
    
  if(f.ARMED) {
	  LEDPIN_ON;	  
  } // 드론이 동작중일 때 시스템 LED 켜기
  else
  {
	  if((calibratingA>0 ) || (calibratingG>0) || (calibratingB>0))
	  {
		//f.ACC_CALIBRATED = 0;
		if ( currentTime > calibratedAccTime )
		{
			LEDPIN_TOGGLE;
			calibratedAccTime = currentTime + 20000;
		}
		else;
		//USB_SerialWrite('#');
	  }
	  else 
	  {
		  if(f.SMALL_ANGLES_25) 
		  {
			  
			if ( currentTime > calibratedAccTime )
			{
				LEDPIN_TOGGLE;
				calibratedAccTime = currentTime + 100000;
			}
			else;
		  } 
		  else if(Drone_State_Flag == COMM_READY || Drone_State_Flag == FAIL_SAFE)
		  {
			  
			if ( currentTime > calibratedAccTime )
			{
				LEDPIN_TOGGLE;
				calibratedAccTime = currentTime + 400000;
			}
			else;
		  } 
		  else LEDPIN_OFF; // 센서 조정완료 후 시스템 LED 끄기
	  }
  }
}

void CodingDrone::setTrim(int x, int y){
    conf.angleTrim[0] = ROLL_TRIM - constrain(x,-100,100); 
	conf.angleTrim[1] = PITCH_TRIM - constrain(y,-100,100);
}
/*
void POSITION_CONTROL()
{
	
	//if(f.ARMED) { if(Alt_Throttle_Offset < 200) Alt_Throttle_Offset++; }
	//else Alt_Throttle_Offset = 0;
	
	// MPU6050
    
}*/

/////////////////////////////////////////////////////////////////////////

void Read_2Byte_Data(uint8_t *A,uint8_t *B,uint8_t size) {
	while(size--) *A++ = *B++;
}

void USB_SerialOpen(uint32_t baud) {
    uint8_t h = ((F_CPU  / 4 / baud - 1) / 2) >> 8;
    uint8_t l = ((F_CPU  / 4 / baud - 1) / 2);
	UDIEN &= ~(1 << SOFE); // disable the USB frame interrupt of arduino (it causes strong jitter and we dont need it)
}

void BT_SerialOpen(uint32_t baud) {
    uint8_t h = ((F_CPU  / 4 / baud - 1) / 2) >> 8;
    uint8_t l = ((F_CPU  / 4 / baud - 1) / 2);
    UCSR1A  = (1 << U2X1); 
	UBRR1H = h; 
	UBRR1L = l; 
	UCSR1B |= (1 << RXEN1) | (1 << TXEN1) | (1 << RXCIE1);
}

void BT_SerialEnd() {
	UCSR1B &= ~((1 << RXEN1) | (1 << TXEN1) | (1 << RXCIE1) | (1 << UDRIE1));
}

bool CodingDrone::BT_Available() {
	if(serialHeadRX[1] != serialTailRX[1]){
		return 1;	
	}
	else return 0;
}

// we don't care about ring buffer overflow (head->tail) to avoid a test condition : data is lost anyway if it happens
void store_uart_in_buf(uint8_t data, uint8_t portnum) {
    uint8_t h = serialHeadRX[portnum];
    serialBufferRX[h++][portnum] = data;
    if (h >= RX_BUFFER_SIZE)  h = 0;
    serialHeadRX[portnum] = h;
}
  

uint8_t CodingDrone::BT_SerialRead() {
	
    uint8_t t = serialTailRX[1];
    uint8_t c = serialBufferRX[t][1];
    if (serialHeadRX[1] != t) {
      if (++t >= RX_BUFFER_SIZE) t = 0;
      serialTailRX[1] = t;
    }
    return c;
}

uint8_t USB_SerialRead() {
	
#if (ARDUINO >= 100)
    USB_Flush(USB_CDC_TX);
#else
    return USB_Recv(USB_CDC_RX);
#endif

    uint8_t t = serialTailRX[0];
    uint8_t c = serialBufferRX[t][0];
    if (serialHeadRX[0] != t) {
      if (++t >= RX_BUFFER_SIZE) t = 0;
      serialTailRX[0] = t;
    }
    return c;
}

void SerialSerialize(uint8_t port, uint8_t a) {
    uint8_t t = serialHeadTX[port];
    if (++t >= TX_BUFFER_SIZE) t = 0;
    serialBufferTX[t][port] = a;
    serialHeadTX[port] = t;
}

static void serialize16(uint8_t port,int16_t a) {
  SerialSerialize(port,(a   ) & 0xFF);
  SerialSerialize(port,(a>>8) & 0xFF);
}
  
void UartSendData(uint8_t port) {
	uint8_t t;
	static uint32_t nowTime = micros();
	static uint16_t periodTime = micros() + 80;
	
    switch (port) {
      case 0:
        while (serialHeadTX[0] != serialTailTX[0]) {
          if (++serialTailTX[0] >= TX_BUFFER_SIZE) serialTailTX[0] = 0;
		  USB_Send(USB_CDC_TX, serialBufferTX[serialTailTX[0]], 1);
        }
        break;
      case 1: //UCSR1B |= (1 << UDRIE1); break;
	  
	    t= serialTailTX[1];	
		while(t != serialHeadTX[1]) {
			while (1) {
				nowTime = micros();
				if ((int16_t)(nowTime - periodTime) > 0) {
					periodTime = nowTime + 80;
					break;
				}
			}
			if (serialHeadTX[1] != t) {
			  if (++t >= TX_BUFFER_SIZE) t = 0;
			  UDR1 = serialBufferTX[t][1];  // Transmit next byte in the ring
			  //UCSR1B &= ~(1 << UDRIE1);
			  serialTailTX[1] = t;
			}
			//delayMicroseconds(80);
		}
		break;
    }
}

void USB_SerialWrite(uint8_t c) {
    SerialSerialize(0, c); UartSendData(0);
}



void CodingDrone::BT_SerialWrite(uint8_t c) {
    SerialSerialize(1, c); UartSendData(1);
}

void CodingDrone::BT_SerialWriteString(String s) {
	
	int t = 0;
	
	for(t = 0; t < s.length();t++){
		SerialSerialize(1, s[t]); 
	}
	UartSendData(1);
	
}

void BT_SerialWrite16(uint16_t c) {
    serialize16(1, c); UartSendData(1);
}


uint8_t SerialAvailable(uint8_t port) {
    if (port == 0) return USB_Available(USB_CDC_RX);
    return ((uint8_t)(serialHeadRX[port] - serialTailRX[port])) % RX_BUFFER_SIZE;
}


ISR(USART1_RX_vect)  {
    store_uart_in_buf(UDR1, 1);
}
/*
ISR(USART1_UDRE_vect) { // Serial 1 on a MEGA or on a PROMICRO
    uint8_t t = serialTailTX[1];
    if (serialHeadTX[1] != t) {
      if (++t >= TX_BUFFER_SIZE) t = 0;
      UDR1 = serialBufferTX[t][1];  // Transmit next byte in the ring
      serialTailTX[1] = t;
    }
    if (t == serialHeadTX[1]) UCSR1B &= ~(1 << UDRIE1);
}
*/

//// CodingDrone_AutoMove Class Function
