#ifndef DEF_H_
#define DEF_H_

/*** RC Controller Middle Position : 대부분의 드론 RC 조종기들은 조이스틱의 중간 위치가 1500입니다 ***/
#define MIDRC 1500

#define MINCHECK 1100
#define MAXCHECK 1900    


/*** SERIAL MONITOR : 시리얼모니터로 데이터를 받고싶으면 아래의 주석을 해제하세요 ***/
//#define Serial_Monitor_Activate
//#define Wireless_Monitor_Activate

/**************************  atmega32u4 (Promicro)  ***********************************/
  
  #define LEDPIN_PINMODE             //
  #define LEDPIN_TOGGLE              PIND |= 1<<5;     //switch LEDPIN state (Port D5)
  #define LEDPIN_ON                PORTD &= ~(1<<5);
  #define LEDPIN_OFF                 PORTD |= (1<<5);
 
  #define I2C_PULLUPS_ENABLE         PORTD |= 1<<0; PORTD |= 1<<1;   // PIN 2&3 (SDA&SCL)
  #define I2C_PULLUPS_DISABLE        PORTD &= ~(1<<0); PORTD &= ~(1<<1);
  
  
  #define RX_SERIAL_PORT           	1
  #define USB_CDC_TX                3
  #define USB_CDC_RX                2
  
  #define V_BATPIN                  A5    // Analog PIN 0

  /**************************************************************************************/
  /***********************                  TX-related         **************************/
  /**************************************************************************************/

    /* introduce a deadband around the stick center
       Must be greater than zero, comment if you dont want a deadband on roll, pitch and yaw */
    #define DEADBAND 10
	
/**************************************************************************************/
/***************      IMU Orientations and Sensor definitions      ********************/
/**************************************************************************************/


#define ACC_ORIENTATION(X, Y, Z)  {imu.accADC[0]  = -X; imu.accADC[1]  = -Y; imu.accADC[2]  =  Z;} 
#define GYRO_ORIENTATION(X, Y, Z) {imu.gyroADC[0] =  Y; imu.gyroADC[1] = -X; imu.gyroADC[2] = -Z;} 
#define MAG_ORIENTATION(X, Y, Z)  {imu.magADC[0]  =  X; imu.magADC[1]  =  Y; imu.magADC[2]  = -Z;} 
//#undef INTERNAL_I2C_PULLUPS
#define MINTHROTTLE 1100
#define MAXTHROTTLE 1950
#define SERIAL_RECEIVER_ONLY  1
#define RC_CHANS 8

/****************************    Mincommand          *******************************/
/* this is the value for the ESCs when they are not armed
   in some cases, this value must be lowered down to 900 for some specific ESCs, otherwise they failed to initiate */
#define MINCOMMAND  1000
#define STANBYCOMMAND 1100

/**********************************  I2C speed for old WMP config (useless config for other sensors)  *************/
//#define I2C_SPEED 100000L     //100kHz normal mode, this value must be used for a genuine WMP
#define I2C_SPEED 400000L   //400kHz fast mode, it works only with some WMP clones

/***************************    Internal i2c Pullups   ********************************/
/* enable internal I2C pull ups (in most cases it is better to use external pullups) */
//#define INTERNAL_I2C_PULLUPS



/********                          Failsafe settings                 ********************/
/* Failsafe check pulses on four main control channels CH1-CH4. If the pulse is missing or bellow 985us (on any of these four channels) 
   the failsafe procedure is initiated. After FAILSAFE_DELAY time from failsafe detection, the level mode is on (if ACC is avaliable),
   PITCH, ROLL and YAW is centered and THROTTLE is set to FAILSAFE_THROTTLE value. You must set this value to descending about 1m/s or so
   for best results. This value is depended from your configuration, AUW and some other params.  Next, after FAILSAFE_OFF_DELAY the copter is disarmed, 
   and motors is stopped. If RC pulse coming back before reached FAILSAFE_OFF_DELAY time, after the small quard time the RC control is returned to normal. */

//#define FAILSAFE                                // uncomment  to activate the failsafe function
#define FAILSAFE_DELAY     10                     // Guard time for failsafe activation after signal lost. 1 step = 0.1sec - 1sec in example
#define FAILSAFE_OFF_DELAY 200                    // Time for Landing before motors stop in 0.1sec. 1 step = 0.1sec - 20sec in example
#define FAILSAFE_THROTTLE  (MINTHROTTLE + 200)    // (*) Throttle level used for landing - may be relative to MINTHROTTLE - as in this case

#define FAILSAFE_DETECT_TRESHOLD  950


/********************************************************************/
/****           battery voltage monitoring                       ****/
/********************************************************************/
/* for V BAT monitoring
   after the resistor divisor we should get [0V;5V]->[0;1023] on analog V_BATPIN
   with R1=33k and R2=51k
   vbat = [0;1023]*16/VBATSCALE
   must be associated with #define BUZZER ! */
#define VBAT              // uncomment this line to activate the vbat code
#define VBATSCALE       129 // (*) (**) change this value if readed Battery voltage is different than real voltage
#define VBATNOMINAL      37 // 3.7V full battery nominal voltage - only used for lcd.telemetry
#define VBATLEVEL_WARN1  30 // (*) (**) 3.0V
#define VBATLEVEL_WARN2  25 // (*) (**) 2.5V
#define VBATLEVEL_CRIT   20 // (*) (**) 2.0V - critical condition: if vbat ever goes below this value, permanent alarm is triggered
#define NO_VBAT          18 // Avoid beeping without any battery
#define VBAT_OFFSET       0.9 // offset in 0.1Volts, gets added to voltage value  - useful for zener diodes
#define VBAT_SMOOTH 64              // len of averaging vector for smoothing the VBAT readings; should be power of 2; set to 1 to disable
#define VBAT_PRESCALER 4 // set this to 8 if vbatscale would exceed 255

/* for V BAT monitoring of individual cells
 * enable both VBAT and VBAT_CELLS
 */
//#define VBAT_CELLS
#define VBAT_CELLS_NUM 0 // set this to the number of cells you monitor via analog pins
#define VBAT_CELLS_PINS {A0, A1, A2, A3, A4, A5 } // set this to the sequence of analog pins
#define VBAT_CELLS_OFFSETS {0, 50, 83, 121, 149, 177 } // in 0.1 volts, gets added to voltage value  - useful for zener diodes
#define VBAT_CELLS_DIVS { 75, 122,  98, 18, 30, 37 } // divisor for proportional part according to resistors - larger value here gives smaller voltage


/********************************************************************/
/****           altitude hold                                    ****/
/********************************************************************/

/* defines the neutral zone of throttle stick during altitude hold, default setting is
   +/-50 uncommend and change the value below if you want to change it. */
#define ALT_HOLD_THROTTLE_NEUTRAL_ZONE    50
#define ALT_HOLD_THROTTLE_MIDPOINT        1500  // in us    - if uncommented, this value is used in ALT_HOLD for throttle stick middle point instead of initialThrottleHold parameter.


/************************        Angele throttle correction         ********************/
  /* Automatically increase throttle based on the angle of the copter
     Original idea by Kraut Rob, first implementation HAdrian */

//#define THROTTLE_ANGLE_CORRECTION 10
//#define THROTTLE_ANGLE_ALTITUDE_COMP


#define ROLL_TRIM -23
#define PITCH_TRIM -42
  
  
#define NUMBER_MOTOR 4

 

#endif /* DEF_H_ */
